@@ -1,36 +1,58 @@
 """Functions for more advanced calculations of quantities and properties of
 quantum objects.
 """
-import numbers
-import itertools
-import functools
+
 import collections
-from math import sin, cos, pi, log, log2, sqrt
+import functools
+import itertools
+import numbers
+from math import cos, log, log2, pi, sin, sqrt
 
 import numpy as np
 import numpy.linalg as nla
+from cotengra import array_contract
 from scipy.optimize import minimize
 
 from .core import (
-    njit, issparse, isop, zeroify, realify, prod, isvec, dot, dag,
-    qu, kron, eye, ikron, tr, ptr, infer_size, expec, dop, ensure_qarray,
-)
-from .linalg.base_linalg import (
-    eigh, eigvalsh, norm, sqrtm,
-)
-from .linalg.approx_spectral import (
-    entropy_subsys_approx, tr_sqrt_subsys_approx,
-    logneg_subsys_approx, gen_bipartite_spectral_fn,
+    dag,
+    dop,
+    dot,
+    ensure_qarray,
+    expec,
+    eye,
+    ikron,
+    infer_size,
+    isop,
+    issparse,
+    isvec,
+    kron,
+    njit,
+    prod,
+    ptr,
+    qu,
+    realify,
+    tr,
+    zeroify,
 )
 from .gen.operators import pauli
-from .gen.states import (
-    basis_vec, bell_state, bloch_state,
+from .gen.states import basis_vec, bell_state, bloch_state
+from .linalg.approx_spectral import (
+    entropy_subsys_approx,
+    gen_bipartite_spectral_fn,
+    logneg_subsys_approx,
+    tr_sqrt_subsys_approx,
+)
+from .linalg.base_linalg import eigh, eigvalsh, norm, norm_trace_dense, sqrtm
+from .utils import (
+    frequencies,
+    int2tup,
+    keymap,
 )
-from .utils import int2tup
 
 
-def fidelity(p1, p2):
-    """Fidelity between two quantum states.
+def fidelity(p1, p2, squared=False):
+    """Fidelity between two quantum states. By default, the unsquared fidelity
+    is used, equivalent in the pure state case to ``|<psi|phi>|``.
 
     Parameters
     ----------
@@ -38,17 +60,26 @@ def fidelity(p1, p2):
         First state.
     p2 : vector or operator
         Second state.
+    squared : bool, optional
+        Whether to use the squared convention or not, by default False.
 
     Returns
     -------
     float
     """
     if isvec(p1) or isvec(p2):
-        return expec(p1, p2)
-    else:
-        sqrho = sqrtm(p1)
-        return tr(sqrtm(dot(sqrho, dot(p2, sqrho))))
-        # return norm(sqrtm(p1) @ sqrtm(p2), "tr")
+        F = expec(p1, p2).real
+        if not squared:
+            F = max(0.0, F) ** 0.5
+        return F
+
+    sqrho = sqrtm(p1)
+    F = tr(sqrtm(dot(sqrho, dot(p2, sqrho)))).real
+
+    if squared:
+        F = F**2
+
+    return F
 
 
 def purify(rho):
@@ -74,6 +105,281 @@ def purify(rho):
     return qu(psi)
 
 
+@ensure_qarray
+def kraus_op(rho, Ek, dims=None, where=None, check=False):
+    r"""Operate on a mixed state with a set of kraus operators:
+
+    .. math::
+
+        \sigma = \sum_k E_k \rho E_k^{\dagger}
+
+    Parameters
+    ----------
+    rho : (d, d) density matrix
+        The density operator to perform operation on.
+    Ek : (K, d, d) array or sequence of K (d, d) arrays
+        The Kraus operator(s).
+    dims : sequence of int, optional
+        The subdimensions of ``rho``.
+    where : int or sequence of int, optional
+        Which of the subsystems to apply the operation on.
+    check : bool, optional
+        Where to check ``sum_k(Ek.H @ Ek) == 1``.
+
+    Returns
+    -------
+    sigma : density matrix
+        The state after the kraus operation.
+
+    Examples
+    --------
+    The depolarising channel:
+
+    .. code-block:: python3
+
+        In [1]: import quimb as qu
+
+        In [2]: rho = qu.rand_rho(2)
+
+        In [3]: I, X, Y, Z = (qu.pauli(s) for s in 'IXYZ')
+
+        In [4]: [qu.expec(rho, A) for A in (X, Y, Z)]
+        Out[4]: [-0.652176185230622, -0.1301762132792484, 0.22362918368272583]
+
+        In [5]: p = 0.1
+
+        In [6]: Ek = [
+           ...:     (1 - p)**0.5 * I,
+           ...:     (p / 3)**0.5 * X,
+           ...:     (p / 3)**0.5 * Y,
+           ...:     (p / 3)**0.5 * Z
+           ...: ]
+
+        In [7]: sigma = qu.kraus_op(rho, Ek)
+
+        In [8]: [qu.expec(sigma, A) for A in (X, Y, Z)]
+        Out[8]: [-0.5652193605332058, -0.11281938484201527, 0.1938119591916957]
+    """
+    if not isinstance(Ek, np.ndarray):
+        Ek = np.stack(Ek, axis=0)
+
+    if check:
+        SEk = np.einsum("kij,kil", Ek.conj(), Ek)
+        if norm(SEk - eye(Ek.shape[-1]), "fro") > 1e-12:
+            raise ValueError("Did not find ``sum(E_k.H @ Ek) == 1``.")
+
+    if int(dims is None) + int(where is None) == 1:
+        raise ValueError("If `dims` is specified so should `where`.")
+
+    if isinstance(where, numbers.Integral):
+        where = (where,)
+
+    if dims:
+        # the full subdimensions of ``rho``
+        dims = tuple(dims)
+        rho = rho.reshape(dims + dims)
+        N = len(dims)
+
+        # the subdimensions of where the kraus operator should act
+        kdims = tuple(dims[i] for i in where)
+        Ek = Ek.reshape((-1,) + kdims + kdims)
+
+        rho_inds = (
+            *(f"i*{q}" if q in where else f"i{q}" for q in range(N)),
+            *(f"j*{q}" if q in where else f"j{q}" for q in range(N)),
+        )
+        Ei_inds = ("K", *(f"i{q}" for q in where), *(f"i*{q}" for q in where))
+        Ej_inds = ("K", *(f"j{q}" for q in where), *(f"j*{q}" for q in where))
+        out = (*(f"i{q}" for q in range(N)), *(f"j{q}" for q in range(N)))
+    else:
+        Ei_inds = ("K", "i", "i*")
+        rho_inds = ("i*", "j*")
+        Ej_inds = ("K", "j", "j*")
+        out = ("i", "j")
+
+    sigma = array_contract(
+        (Ek, rho, Ek.conj()),
+        (Ei_inds, rho_inds, Ej_inds),
+        out,
+    )
+
+    if dims:
+        sigma = sigma.reshape(prod(dims), prod(dims))
+
+    return sigma
+
+
+@ensure_qarray
+def projector(A, eigenvalue=1.0, tol=1e-12, autoblock=False):
+    """Compute the projector for the target ``eigenvalue`` of operator
+    ``A``.
+
+    Parameters
+    ----------
+    A : operator or tuple[array, operator].
+        The hermitian observable. If a tuple is supplied, assume this is the
+        eigendecomposition of the observable.
+    eigenvalue : float, optional
+        The target eigenvalue to construct the projector for, default: 1.
+    tol : float, optional
+        The tolerance with which to select all eigenvalues.
+    autoblock : bool, optional
+        Whether to use automatic symmetry exploitation.
+
+    Returns
+    -------
+    P : dense matrix
+        The projector onto the target eigenspace.
+    """
+    if isinstance(A, (tuple, list)):
+        el, ev = A
+    else:
+        el, ev = eigh(A, autoblock=autoblock)
+    which = np.argwhere(abs(el - eigenvalue) < tol)
+    P = np.zeros_like(ev)
+    for i in which:
+        vi = ev[:, i]
+        P += vi @ vi.H
+    return P
+
+
+def measure(p, A, eigenvalue=None, tol=1e-12):
+    r"""Measure state ``p`` with observable ``A``, collapsing the state in the
+    process and returning the relevant eigenvalue.
+
+    .. math::
+
+        \left| \psi \right\rangle \rightarrow
+        \frac{
+            P_i \left| \psi \right\rangle
+        }{
+            \sqrt{\left\langle \psi \right| P_i \left| \psi \right\rangle}
+        }
+
+    and
+
+    .. math::
+
+        \rho \rightarrow
+        \frac{
+            P_i \rho P_i^{\dagger}
+        }{
+            \text{Tr} \left[ P_i \rho \right]
+        }
+
+    along with the corresponding eigenvalue :math:`\lambda_i`.
+
+    Parameters
+    ----------
+    p : vector or matrix
+        The quantum state to measure.
+    A : matrix or tuple[array, matrix]
+        The hermitian operator corresponding to an observable. You can supply
+        also a pre-diagonalized operator here as a tuple of eigenvalues and
+        eigenvectors.
+    tol : float, optional
+        The tolerance within which to group eigenspaces.
+    eigenvalue : float, optional
+        If specified, deterministically collapse to this result. Otherwise
+        randomly choose a result as in 'real-life'.
+
+    Returns
+    -------
+    result : float
+        The result of the measurement.
+    p_after : vector or matrix
+        The quantum state post measurement.
+    """
+    if isinstance(A, (tuple, list)):
+        el, ev = A
+    else:
+        el, ev = eigh(A)
+
+    js = np.arange(el.size)
+
+    # compute prob of each eigenvector
+    if isvec(p):
+        pj = (abs(ev.H @ p) ** 2).flatten()
+    else:
+        pj = array_contract(
+            (ev.H, p, ev),
+            ("jk", "kl", "lj"),
+            ("j",),
+        ).real
+
+    # then choose one
+    if eigenvalue is None:
+        j = np.random.choice(js, p=pj)
+        eigenvalue = el[j]
+
+    # now combine whole eigenspace
+    P = projector((el, ev), eigenvalue=eigenvalue, tol=tol)
+    total_prob = np.sum(pj[abs(el - eigenvalue) < tol])
+
+    # now collapse the state
+    if isvec(p):
+        p_after = P @ (p / total_prob**0.5)
+    else:
+        p_after = (P @ p @ P.H) / total_prob
+
+    return eigenvalue, p_after
+
+
+def simulate_counts(p, C, phys_dim=2, seed=None):
+    """Simulate measuring each qubit of ``p`` in the computational basis,
+    producing output like that of ``qiskit``.
+
+    Parameters
+    ----------
+    p : vector or operator
+        The quantum state, assumed to be normalized, as either a ket or density
+        operator.
+    C : int
+        The number of counts to perform.
+    phys_dim : int, optional
+        The assumed size of the subsystems of ``p``, defaults to 2 for qubits.
+
+    Returns
+    -------
+    results : dict[str, int]
+        The counts for each bit string measured.
+
+    Examples
+    --------
+
+    Simulate measuring the state of each qubit in a GHZ-state:
+
+    .. code:: python3
+
+        >>> import quimb as qu
+        >>> psi = qu.ghz_state(3)
+        >>> qu.simulate_counts(psi, 1024)
+        {'000': 514, '111': 510}
+
+    """
+    rng = np.random.default_rng(seed)
+
+    n = infer_size(p, phys_dim)
+    d = phys_dim**n
+
+    if isop(p):
+        pi = np.diag(p).real
+    else:
+        pi = np.multiply(np.conj(p), p).real
+
+    # probability of each basis state
+    pi = pi.reshape(-1)
+
+    # raw counts in terms of integers
+    raw_counts = rng.choice(d, size=C, p=pi)
+
+    # convert to frequencies of binary
+    bin_str = "{:0>" + str(n) + "b}"
+    results = keymap(bin_str.format, frequencies(raw_counts))
+
+    return results
+
+
 def dephase(rho, p, rand_rank=None):
     """Dephase ``rho`` by amount ``p``, that is, mix it
     with the maximally mixed state:
@@ -141,7 +447,7 @@ def entropy(a, rank=None):
         if rank is None:
             evals = eigvalsh(a)
         else:  # know that not all eigenvalues needed
-            evals = eigvalsh(a, k=rank, which='LM', backend='AUTO')
+            evals = eigvalsh(a, k=rank, which="LM", backend="AUTO")
 
     evals = evals[evals > 0.0]
     return np.sum(-evals * np.log2(evals))
@@ -241,12 +547,15 @@ def check_dims_and_indices(dims, *syss):
     all_sys = sum(syss, ())
 
     if not all(0 <= i < nsys for i in all_sys):
-        raise ValueError("Indices specified in `sysa` and `sysb` must be "
-                         "in range({}) for dims {}.".format(nsys, dims))
+        raise ValueError(
+            "Indices specified in `sysa` and `sysb` must be "
+            f"in range({nsys}) for dims {dims}."
+        )
 
 
-def mutinf_subsys(psi_abc, dims, sysa, sysb, approx_thresh=2**13,
-                  **approx_opts):
+def mutinf_subsys(
+    psi_abc, dims, sysa, sysb, approx_thresh=2**13, **approx_opts
+):
     """Calculate the mutual information of two subsystems of a pure state,
     possibly using an approximate lanczos method for large subsytems.
 
@@ -283,7 +592,7 @@ def mutinf_subsys(psi_abc, dims, sysa, sysb, approx_thresh=2**13,
     sz_b = prod(d for i, d in enumerate(dims) if i in sysb)
     sz_c = prod(dims) // (sz_a * sz_b)
 
-    kws = {'approx_thresh': approx_thresh, **approx_opts}
+    kws = {"approx_thresh": approx_thresh, **approx_opts}
 
     if sz_c == 1:
         hab = 0.0
@@ -329,17 +638,16 @@ def schmidt_gap(psi_ab, dims, sysa):
         sysa = sysb
 
     rho_a = ptr(psi_ab, dims, sysa)
-    el = eigvalsh(rho_a, k=2, which='LM')
+    el = eigvalsh(rho_a, k=2, which="LM")
     return abs(el[0] - el[1])
 
 
 def tr_sqrt(A, rank=None):
-    """Return the trace of the sqrt of a positive semidefinite operator.
-    """
+    """Return the trace of the sqrt of a positive semidefinite operator."""
     if rank is None:
         el = eigvalsh(A, sort=False)
     else:
-        el = eigvalsh(A, k=rank, which='LM', backend='AUTO')
+        el = eigvalsh(A, k=rank, which="LM", backend="AUTO")
     return np.sum(np.sqrt(el[el > 0.0]))
 
 
@@ -407,10 +715,12 @@ def partial_transpose(p, dims=(2, 2), sysa=0):
             perm_ket_inds.append(i)
             perm_bra_inds.append(i + ndims)
 
-    return (np.asarray(qu(p, "dop"))
-            .reshape((*dims, *dims))
-            .transpose((*perm_ket_inds, *perm_bra_inds))
-            .reshape((prod(dims), prod(dims))))
+    return (
+        np.asarray(qu(p, "dop"))
+        .reshape((*dims, *dims))
+        .transpose((*perm_ket_inds, *perm_bra_inds))
+        .reshape((prod(dims), prod(dims)))
+    )
 
 
 def partial_transpose_norm(p, dims, sysa):
@@ -432,9 +742,9 @@ def partial_transpose_norm(p, dims, sysa):
             sysa = sysb
 
         rhoa = ptr(p, dims, sysa)
-        return tr_sqrt(rhoa)**2
+        return tr_sqrt(rhoa) ** 2
 
-    return norm(partial_transpose(p, dims, sysa), "tr")
+    return norm_trace_dense(partial_transpose(p, dims, sysa), isherm=True)
 
 
 @zeroify
@@ -467,8 +777,9 @@ def logneg(p, dims=(2, 2), sysa=0):
 logarithmic_negativity = logneg
 
 
-def logneg_subsys(psi_abc, dims, sysa, sysb,
-                  approx_thresh=2**13, **approx_opts):
+def logneg_subsys(
+    psi_abc, dims, sysa, sysb, approx_thresh=2**13, **approx_opts
+):
     """Compute the logarithmic negativity between two subsystems of a pure
     state, possibly using an approximate lanczos for large subsystems. Uses
     a special method if the two subsystems form a bipartition of the state.
@@ -509,8 +820,12 @@ def logneg_subsys(psi_abc, dims, sysa, sysb,
 
     # check for pure bipartition
     if sz_c == 1:
-        psi_ab_ppt_norm = tr_sqrt_subsys(
-            psi_abc, dims, sysa, approx_thresh=approx_thresh, **approx_opts)**2
+        psi_ab_ppt_norm = (
+            tr_sqrt_subsys(
+                psi_abc, dims, sysa, approx_thresh=approx_thresh, **approx_opts
+            )
+            ** 2
+        )
         return max(log2(psi_ab_ppt_norm), 0.0)
 
     # check whether to use approx lanczos method
@@ -585,11 +900,11 @@ def concurrence(p, dims=(2, 2), sysa=0, sysb=1):
     if len(dims) > 2:
         p = ptr(p, dims, (sysa, sysb))
 
-    Y = pauli('Y')
+    Y = pauli("Y")
 
     if isop(p):
         pt = dot(kron(Y, Y), dot(p.conj(), kron(Y, Y)))
-        evals = (nla.eigvals(dot(p, pt)).real**2)**0.25
+        evals = (nla.eigvals(dot(p, pt)).real ** 2) ** 0.25
         return max(0, 2 * np.max(evals) - np.sum(evals))
     else:
         pt = dot(kron(Y, Y), p.conj())
@@ -626,13 +941,22 @@ def one_way_classical_information(p_ab, prjs, precomp_func=False):
                 prob = tr(p_ab_j)
                 p_a_j = ptr(p_ab_j, (2, 2), 0) / prob
                 yield prob, p_a_j
+
         return s_a - sum(p * entropy(rho) for p, rho in gen_paj())
 
     return owci if precomp_func else owci(prjs)
 
 
 @zeroify
-def quantum_discord(p, dims=(2, 2), sysa=0, sysb=1):
+def quantum_discord(
+    p,
+    dims=(2, 2),
+    sysa=0,
+    sysb=1,
+    method="COBYLA",
+    tol=1e-12,
+    maxiter=2**14,
+):
     """Quantum Discord for two qubit density operator.
 
     If ``len(dims) > 2``, then the non-target dimensions will be traced out
@@ -666,8 +990,16 @@ def quantum_discord(p, dims=(2, 2), sysa=0, sysb=1):
         prjb = eye(2) - prja
         return iab - owci((prja, prjb))
 
-    opt = minimize(trial_qd, (pi / 2, pi),
-                   method="SLSQP", bounds=((0, pi), (0, 2 * pi)))
+    opt = minimize(
+        trial_qd,
+        (pi / 2, pi),
+        method=method,
+        bounds=((0, pi), (0, 2 * pi)),
+        tol=tol,
+        options=dict(
+            maxiter=maxiter,
+        ),
+    )
     if opt.success:
         return opt.fun
     else:  # pragma: no cover
@@ -676,7 +1008,22 @@ def quantum_discord(p, dims=(2, 2), sysa=0, sysb=1):
 
 @zeroify
 def trace_distance(p1, p2):
-    """Trace distance between two states.
+    r"""Trace distance between two states:
+
+    .. math::
+
+        \delta(\rho, \sigma)
+        =
+        \frac{1}{2} \left| \rho - \sigma \right|_\mathrm{tr}
+
+    If two wavefunctions are supplied the trace distance will be computed via
+    the more efficient expression:
+
+    .. math::
+
+        \delta(|\psi\rangle\langle\psi|, |\phi\rangle\langle\phi|)
+        =
+        \sqrt{1 - \langle \psi | \phi \rangle^2}
 
     Parameters
     ----------
@@ -696,8 +1043,56 @@ def trace_distance(p1, p2):
         return sqrt(1 - expec(p1, p2))
 
     # Otherwise do full calculation
-    return 0.5 * norm((p1 if p1_is_op else dop(p1)) -
-                      (p2 if p2_is_op else dop(p2)), "tr")
+    return 0.5 * norm(
+        (p1 if p1_is_op else dop(p1)) - (p2 if p2_is_op else dop(p2)), "tr"
+    )
+
+
+def cprint(psi, prec=6):
+    """Print a state in the computational basis.
+
+    Parameters
+    ----------
+    psi : vector
+        The pure state.
+    prec : int, optional
+        How many significant figures to print.
+
+    Examples
+    --------
+
+        >>> cprint(rand_ket(2**2))
+        (-0.0751069-0.192635j) |00>
+          (0.156837-0.562213j) |01>
+        (-0.307381+0.0291168j) |10>
+          (-0.14302+0.707661j) |11>
+
+        >>> cprint(w_state(4))
+        (0.5+0j) |0001>
+        (0.5+0j) |0010>
+        (0.5+0j) |0100>
+        (0.5+0j) |1000>
+    """
+    d = psi.shape[0]
+    n = int(log2(d))
+    if 2**n != d:
+        raise ValueError("State is not factorizable into computational basis.")
+
+    coeff_str = "{:." + str(prec) + "}"
+
+    coeffs, cs = [], []
+    for c in itertools.product("01", repeat=n):
+        c = "".join(c)
+        ix = int(c, 2)
+        coeff = psi.item(ix)
+        if coeff != 0.0:
+            cs.append(c)
+            coeffs.append(coeff_str.format(coeff))
+
+    max_str_len = max(map(len, coeffs))
+    full_str = "{:>" + str(max_str_len) + "} |{}>"
+    for coeff, c in zip(coeffs, cs):
+        print(full_str.format(coeff, c))
 
 
 def decomp(a, fn, fn_args, fn_d, nmlz_func, mode="p", tol=1e-3):
@@ -760,23 +1155,20 @@ def decomp(a, fn, fn_args, fn_d, nmlz_func, mode="p", tol=1e-3):
         return names_cffs
 
 
-pauli_decomp = functools.partial(decomp,
-                                 fn=pauli,
-                                 fn_args='IXYZ',
-                                 fn_d=2,
-                                 nmlz_func=lambda n: 2**-n)
+pauli_decomp = functools.partial(
+    decomp, fn=pauli, fn_args="IXYZ", fn_d=2, nmlz_func=lambda n: 2**-n
+)
 """Decompose an operator into Paulis."""
 
-bell_decomp = functools.partial(decomp,
-                                fn=bell_state,
-                                fn_args=(0, 1, 2, 3),
-                                fn_d=4,
-                                nmlz_func=lambda x: 1)
+bell_decomp = functools.partial(
+    decomp, fn=bell_state, fn_args=(0, 1, 2, 3), fn_d=4, nmlz_func=lambda x: 1
+)
 """Decompose an operator into bell-states."""
 
 
-def correlation(p, A, B, sysa, sysb, dims=None, sparse=None,
-                precomp_func=False):
+def correlation(
+    p, A, B, sysa, sysb, dims=None, sparse=None, precomp_func=False
+):
     """Calculate the correlation between two sites given two operators.
 
     Parameters
@@ -812,9 +1204,11 @@ def correlation(p, A, B, sysa, sysb, dims=None, sparse=None,
     if sparse is None:
         sparse = issparse(A) or issparse(B)
 
-    opts = {'sparse': sparse,
-            'coo_build': sparse,
-            'stype': 'csr' if sparse else None}
+    opts = {
+        "sparse": sparse,
+        "coo_build": sparse,
+        "stype": "csr" if sparse else None,
+    }
     opab = ikron((A, B), dims, (sysa, sysb), **opts)
     A = ikron((A,), dims, sysa, **opts)
     B = ikron((B,), dims, sysb, **opts)
@@ -826,8 +1220,9 @@ def correlation(p, A, B, sysa, sysb, dims=None, sparse=None,
     return corr if precomp_func else corr(p)
 
 
-def pauli_correlations(p, ss=("xx", "yy", "zz"), sysa=0, sysb=1,
-                       sum_abs=False, precomp_func=False):
+def pauli_correlations(
+    p, ss=("xx", "yy", "zz"), sysa=0, sysb=1, sum_abs=False, precomp_func=False
+):
     """Calculate the correlation between sites for a list of operator pairs
     choisen from the pauli matrices.
 
@@ -854,13 +1249,14 @@ def pauli_correlations(p, ss=("xx", "yy", "zz"), sysa=0, sysb=1,
         the correlations for an arbitrary state, depending on ``sum_abs`` and
         ``precomp_func``.
     """
+
     def gen_corr_list():
         for s1, s2 in ss:
-            yield correlation(p, pauli(s1), pauli(s2), sysa, sysb,
-                              precomp_func=precomp_func)
+            yield correlation(
+                p, pauli(s1), pauli(s2), sysa, sysb, precomp_func=precomp_func
+            )
 
     if sum_abs:
-
         if precomp_func:
             return lambda p: sum((abs(corr(p)) for corr in gen_corr_list()))
 
@@ -869,8 +1265,9 @@ def pauli_correlations(p, ss=("xx", "yy", "zz"), sysa=0, sysb=1,
     return tuple(gen_corr_list())
 
 
-def ent_cross_matrix(p, sz_blc=1, ent_fn=logneg, calc_self_ent=True,
-                     upscale=False):
+def ent_cross_matrix(
+    p, sz_blc=1, ent_fn=logneg, calc_self_ent=True, upscale=False
+):
     """Calculate the pair-wise function ent_fn  between all sites or blocks
     of a state.
 
@@ -919,13 +1316,18 @@ def ent_cross_matrix(p, sz_blc=1, ent_fn=logneg, calc_self_ent=True,
                     if calc_self_ent:
                         rhoa = ptr(p, dims, [i + b for b in range(sz_blc)])
                         psiap = purify(rhoa)
-                        ent = ent_fn(psiap,
-                                     dims=(2**sz_blc, 2**sz_blc)) / sz_blc
+                        ent = (
+                            ent_fn(psiap, dims=(2**sz_blc, 2**sz_blc)) / sz_blc
+                        )
                     else:
                         ent = np.nan
                 else:
-                    rhoab = ptr(p, dims, [i + b for b in range(sz_blc)] +
-                                         [j + b for b in range(sz_blc)])
+                    rhoab = ptr(
+                        p,
+                        dims,
+                        [i + b for b in range(sz_blc)]
+                        + [j + b for b in range(sz_blc)],
+                    )
                     ent = ent_fn(rhoab, dims=(2**sz_blc, 2**sz_blc)) / sz_blc
                 ents[i // sz_blc, j // sz_blc] = ent
                 ents[j // sz_blc, i // sz_blc] = ent
@@ -935,33 +1337,50 @@ def ent_cross_matrix(p, sz_blc=1, ent_fn=logneg, calc_self_ent=True,
 
         for i in range(n):
             for j in range(i, n):
-                up_ents[i * sz_blc:(i + 1) * sz_blc,
-                        j * sz_blc:(j + 1) * sz_blc] = ents[i, j]
-                up_ents[j * sz_blc:(j + 1) * sz_blc,
-                        i * sz_blc:(i + 1) * sz_blc] = ents[j, i]
+                up_ents[
+                    i * sz_blc : (i + 1) * sz_blc,
+                    j * sz_blc : (j + 1) * sz_blc,
+                ] = ents[i, j]
+                up_ents[
+                    j * sz_blc : (j + 1) * sz_blc,
+                    i * sz_blc : (i + 1) * sz_blc,
+                ] = ents[j, i]
 
         ents = up_ents
 
     return ents
 
 
-def qid(p, dims, inds, precomp_func=False, sparse_comp=True,
-        norm_func=norm, power=2, coeff=1):
+def qid(
+    p,
+    dims,
+    inds,
+    precomp_func=False,
+    sparse_comp=True,
+    norm_func=norm,
+    power=2,
+    coeff=1,
+):
     # Check inputs
     inds = (inds,) if isinstance(inds, numbers.Number) else inds
 
     # Construct operators
-    ops_i = tuple(tuple(ikron(pauli(s), dims, ind, sparse=sparse_comp)
-                        for s in "xyz")
-                  for ind in inds)
+    ops_i = tuple(
+        tuple(ikron(pauli(s), dims, ind, sparse=sparse_comp) for s in "xyz")
+        for ind in inds
+    )
 
     # Define function closed over precomputed operators
     def qid_func(x):
         if isvec(x):
             x = dop(x)
-        return tuple(sum(coeff * norm_func(dot(x, op) - dot(op, x))**power
-                         for op in ops)
-                     for ops in ops_i)
+        return tuple(
+            sum(
+                coeff * norm_func(dot(x, op) - dot(op, x)) ** power
+                for op in ops
+            )
+            for ops in ops_i
+        )
 
     return qid_func if precomp_func else qid_func(p)
 
@@ -1068,5 +1487,5 @@ def heisenberg_energy(L):
     """
     Einf = (0.5 - 2 * log(2)) * L
     Efinite = pi**2 / (6 * L)
-    correction = 1 + 0.375 / log(L)**3
+    correction = 1 + 0.375 / log(L) ** 3
     return (Einf - Efinite * correction) / 2
