@@ -9,11 +9,27 @@ import functools
 
 import numpy as np
 from scipy.integrate import complex_ode
+from scipy.sparse.linalg import LinearOperator
 
-from .core import (qarray, isop, ldmul, rdmul, explt,
-                   dot, issparse, qu, eye, dag)
-from .linalg.base_linalg import eigh, norm, expm_multiply
-from .utils import continuous_progbar, progbar
+from .core import (
+    dag,
+    dot,
+    explt,
+    eye,
+    isop,
+    issparse,
+    ldmul,
+    make_immutable,
+    qarray,
+    qu,
+    rdmul,
+)
+from .linalg.base_linalg import eigh, norm, expm_multiply, Lazy
+from .linalg.approx_spectral import norm_fro_approx
+from .utils import continuous_progbar, progbar, ensure_dict
+
+
+CALLABLE_TIME_INDEP_CLASSES = (LinearOperator, Lazy)
 
 
 # --------------------------------------------------------------------------- #
@@ -36,12 +52,34 @@ def schrodinger_eq_ket(ham):
     psi_dot(t, y) : callable
         Function to calculate psi_dot(t) at psi(t).
     """
+
     def psi_dot(_, y):
         return -1.0j * dot(ham, y)
 
     return psi_dot
 
 
+def schrodinger_eq_ket_timedep(ham):
+    """Wavefunction time dependent schrodinger equation.
+
+    Parameters
+    ----------
+    ham : callable
+        Time-dependant Hamiltonian governing evolution, such that ``ham(t)``
+        returns an operator representation of the Hamiltonian at time ``t``.
+
+    Returns
+    -------
+    psi_dot(t, y) : callable
+        Function to calculate psi_dot(t) at psi(t).
+    """
+
+    def psi_dot(t, y):
+        return -1.0j * dot(ham(t), y)
+
+    return psi_dot
+
+
 def schrodinger_eq_dop(ham):
     """Density operator schrodinger equation, but with flattened input/output.
 
@@ -69,6 +107,35 @@ def schrodinger_eq_dop(ham):
     return rho_dot
 
 
+def schrodinger_eq_dop_timedep(ham):
+    """Time dependent density operator schrodinger equation, but with flattened
+    input/output.
+
+    Note that this assumes both `ham(t)` and `rho` are hermitian in order to
+    speed up the commutator, non-hermitian hamiltonians as used to model loss
+    should be treated explicilty or with `schrodinger_eq_dop_vectorized`.
+
+    Parameters
+    ----------
+    ham : callable
+        Time-dependant Hamiltonian governing evolution, such that ``ham(t)``
+        returns an operator representation of the Hamiltonian at time ``t``.
+
+    Returns
+    -------
+    rho_dot(t, y) : callable
+        Function to calculate rho_dot(t) at rho(t), input and
+        output both in ravelled (1D form).
+    """
+    d = ham(0).shape[0]
+
+    def rho_dot(t, y):
+        hrho = dot(ham(t), y.reshape(d, d))
+        return -1.0j * (hrho - hrho.T.conj()).reshape(-1)
+
+    return rho_dot
+
+
 def schrodinger_eq_dop_vectorized(ham):
     """Density operator schrodinger equation, but with flattened input/output
     and vectorised superoperator mode (no reshaping required).
@@ -92,6 +159,7 @@ def schrodinger_eq_dop_vectorized(ham):
 
     def rho_dot(_, y):
         return dot(evo_superop, y)
+
     return rho_dot
 
 
@@ -118,8 +186,9 @@ def lindblad_eq(ham, ls, gamma):
 
     def gen_l_terms(rho):
         for l, ll in zip(ls, lls):
-            yield (dot(l, dot(rho, dag(l))) -
-                   0.5 * (dot(rho, ll) + dot(ll, rho)))
+            yield (
+                dot(l, dot(rho, dag(l))) - 0.5 * (dot(rho, ll) + dot(ll, rho))
+            )
 
     def rho_dot(_, y):
         rho = y.reshape(d, d)
@@ -160,34 +229,68 @@ def lindblad_eq_vectorized(ham, ls, gamma, sparse=False):
         for l in ls:
             lb_sparse = issparse(l) or sparse
             idt = eye(d, sparse=lb_sparse)
-            yield ((l & l.conj()) - 0.5 * ((idt & dot(dag(l), l).T) +
-                                           (dot(dag(l), l) & idt)))
+            yield (
+                (l & l.conj())
+                - 0.5 * ((idt & dot(dag(l), l).T) + (dot(dag(l), l) & idt))
+            )
 
     evo_superop += gamma * sum(gen_lb_terms())
 
     def rho_dot(_, y):
         return dot(evo_superop, y)
+
     return rho_dot
 
 
-def _calc_evo_eq(isdop, issparse, isopen=False):
-    """Choose an appropirate dynamical equation to evolve with.
-    """
+def _calc_evo_eq(isdop, issparse, isopen=False, timedep=False):
+    """Choose an appropirate dynamical equation to evolve with."""
     eq_chooser = {
-        (0, 0, 0): schrodinger_eq_ket,
-        (0, 1, 0): schrodinger_eq_ket,
-        (1, 0, 0): schrodinger_eq_dop,
-        (1, 1, 0): schrodinger_eq_dop_vectorized,
-        (1, 0, 1): lindblad_eq,
-        (1, 1, 1): lindblad_eq_vectorized,
+        (0, 0, 0, 0): schrodinger_eq_ket,
+        (0, 1, 0, 0): schrodinger_eq_ket,
+        (1, 0, 0, 0): schrodinger_eq_dop,
+        (1, 1, 0, 0): schrodinger_eq_dop_vectorized,
+        (1, 0, 1, 0): lindblad_eq,
+        (1, 1, 1, 0): lindblad_eq_vectorized,
+        # time-dependent
+        (0, 0, 0, 1): schrodinger_eq_ket_timedep,
+        (0, 1, 0, 1): schrodinger_eq_ket_timedep,
+        (1, 0, 0, 1): schrodinger_eq_dop_timedep,
+        (1, 1, 0, 1): schrodinger_eq_dop_timedep,
     }
-    return eq_chooser[(isdop, issparse, isopen)]
+    return eq_chooser[(isdop, issparse, isopen, timedep)]
+
+
+class Try2Then3Args:
+    def __init__(self, fn):
+        self.fn = fn
+        self.num_args = None
+
+    def first_call(self, t, p, H):
+        try:
+            res = self.fn(t, p)
+            self.num_args = 2
+        except TypeError as e:
+            if "positional" in e.args[0]:
+                res = self.fn(t, p, H)
+                self.num_args = 3
+            else:
+                raise
+        return res
+
+    def __call__(self, t, p, H):
+        if self.num_args is None:
+            return self.first_call(t, p, H)
+        elif self.num_args == 2:
+            return self.fn(t, p)
+        elif self.num_args == 3:
+            return self.fn(t, p, H)
 
 
 # --------------------------------------------------------------------------- #
 # Quantum Evolution Class                                                     #
 # --------------------------------------------------------------------------- #
 
+
 class Evolution(object):
     """A class for evolving quantum systems according to Schrodinger equation.
 
@@ -203,14 +306,22 @@ class Evolution(object):
     ----------
     p0 : quantum state
         Inital state, either vector or operator. If vector, converted to ket.
-    ham : operator, or tuple (1d array, operator).
+    ham : operator, tuple (1d array, operator), or callable
         Governing Hamiltonian, if tuple then assumed to contain
-        ``(eigvals, eigvecs)`` of presolved system.
+        ``(eigvals, eigvecs)`` of presolved system. If callable (but not a
+        SciPy ``LinearOperator``), assume a time-dependent hamiltonian such
+        that ``ham(t)`` is the Hamiltonian at time ``t``. In this case, the
+        latest call to ``ham`` will be cached (and made immutable) in case it
+        is needed by callbacks passed to ``compute``.
     t0 : float, optional
         Initial time (i.e. time of state ``p0``), defaults to zero.
     compute : callable, or dict of callable, optional
-        Function(s) to compute on the state at each time step, called
-        with args (t, pt). If supplied with:
+        Function(s) to compute on the state at each time step. Function(s)
+        should take args (t, pt) or (t, pt, ham) if the Hamiltonian is
+        required. If ham is required, it will be passed in to the function
+        exactly as given to this ``Evolution`` instance, except if ``method``
+        is ``'solve'``, in which case it will be passed in as the solved system
+        ``(eigvals, eigvecs)``. If supplied with:
 
             - single callable : ``Evolution.results`` will contain the results
               as a list,
@@ -218,6 +329,14 @@ class Evolution(object):
               results as a dict of lists with corresponding keys to those
               given in ``compute``.
 
+    int_stop : callable, optional
+        A condition to terminate the integration early if ``method`` is
+        ``'integrate'``. This callable is called at every successful
+        integration step and should take args (t, pt) or (t, pt, ham) similar
+        to the function(s) in the ``compute`` argument.  It should return
+        ``-1`` to stop the integration, otherwise it should return ``None``
+        or ``0``.
+
     method : {'integrate', 'solve', 'expm'}
         How to evolve the system:
 
@@ -245,128 +364,226 @@ class Evolution(object):
         with the ``update_to`` method.
     """
 
-    def __init__(self, p0, ham, t0=0,
-                 compute=None,
-                 method='integrate',
-                 int_small_step=False,
-                 expm_backend='AUTO',
-                 expm_opts=None,
-                 progbar=False):
-
+    def __init__(
+        self,
+        p0,
+        ham,
+        t0=0,
+        compute=None,
+        int_stop=None,
+        method="integrate",
+        int_small_step=False,
+        expm_backend="AUTO",
+        expm_opts=None,
+        progbar=False,
+    ):
         self._p0 = qu(p0)
         self._t = self.t0 = t0
         self._isdop = isop(self._p0)  # Density operator evolution?
         self._d = p0.shape[0]  # Hilbert space dimension
         self._progbar = progbar
 
-        self._setup_callback(compute)
+        self._timedep = callable(ham) and not isinstance(
+            ham, CALLABLE_TIME_INDEP_CLASSES
+        )
+
+        if self._timedep:
+            # cache the time-dependent Hamiltonian in case callbacks use it
+            noncacheing_ham = ham
+
+            @functools.lru_cache(1)
+            def ham(t):
+                Ht = noncacheing_ham(t)
+                if not isinstance(Ht, LinearOperator):
+                    make_immutable(Ht)
+                return Ht
+
+        if (int_stop is not None) and (method != "integrate"):
+            raise ValueError(
+                "You can't provide an integration stopping "
+                "condition (int_stop) if the method is not "
+                "'integrate'"
+            )
+
+        self._setup_callback(compute, int_stop)
         self._method = method
 
-        if method == 'solve' or isinstance(ham, (tuple, list)):
-            self._solve_ham(ham)
-        elif method == 'integrate':
+        if method == "solve" or isinstance(ham, (tuple, list)):
+            if isinstance(ham, LinearOperator):
+                raise TypeError(
+                    "You can't use the 'solve' method "
+                    "with an abstract linear operator "
+                    "Hamiltonian."
+                )
+            elif self._timedep:
+                raise TypeError(
+                    "You can't use the 'solve' method "
+                    "with a time-dependent Hamiltonian."
+                )
+            self._ham = ham
+            self._setup_solved_ham()
+
+        elif method == "integrate":
             self._start_integrator(ham, int_small_step)
-        elif method == 'expm':
+            self._ham = ham
+        elif method == "expm":
+            if isinstance(ham, LinearOperator):
+                raise TypeError(
+                    "You can't use the 'expm' method "
+                    "with an abstract linear operator "
+                    "Hamiltonian."
+                )
+            elif self._timedep:
+                raise TypeError(
+                    "You can't use the 'expm' method "
+                    "with a time-dependent Hamiltonian."
+                )
             self._update_method = self._update_to_expm_ket
             self._pt = self._p0
-            self.ham = ham
             self.expm_backend = expm_backend
-            self.expm_opts = {} if expm_opts is None else dict(expm_opts)
+            self.expm_opts = ensure_dict(expm_opts)
+            self._ham = ham
         else:
-            raise ValueError("Did not understand evolution method: '{}'."
-                             .format(method))
+            raise ValueError(
+                f"Did not understand evolution method: '{method}'."
+            )
 
-    def _setup_callback(self, fn):
-        """Setup callbacks in the correct place to compute into _results
-        """
+    def _setup_callback(self, fn, int_stop):
+        """Setup callbacks in the correct place to compute into _results"""
+        # if fn is None there is no callback
         if fn is None:
             step_callback = None
-            if not self._progbar:
-                int_step_callback = None
+        # else fn is a dict of callbacks or a single callback
+        else:
+            # dict of funcs input -> dict of funcs output
+            if isinstance(fn, dict):
+                fn_try2then3args = {k: Try2Then3Args(v) for k, v in fn.items()}
+                self._results = {k: [] for k in fn}
+
+                def step_callback(t, pt, H):
+                    for k, v in fn_try2then3args.items():
+                        fn_result = v(t, pt, H)
+                        self._results[k].append(fn_result)
+
+            # else results -> single list of outputs of fn
             else:
-                def int_step_callback(t, y):
-                    pass
+                fn_try2then3args = Try2Then3Args(fn)
+                self._results = []
 
-        # dict of funcs input -> dict of funcs output
-        elif isinstance(fn, dict):
-            self._results = {k: [] for k in fn}
-
-            @functools.wraps(fn)
-            def step_callback(t, pt):
-                for k, v in fn.items():
-                    self._results[k].append(v(t, pt))
+                def step_callback(t, pt, H):
+                    fn_result = fn_try2then3args(t, pt, H)
+                    self._results.append(fn_result)
 
             # For the integration callback, additionally need to convert
             #   back to 'quantum' (column vector) form
-            @functools.wraps(fn)
-            def int_step_callback(t, y):
-                pt = qarray(y.reshape(self._d, -1))
-                for k, v in fn.items():
-                    self._results[k].append(v(t, pt))
 
-        # else results -> single list of outputs of fn
+        # if no compute callback, check if there is an int_stop callback
+        if step_callback is None:
+            # if there is only an int_stop callback, set this up
+            if int_stop is not None:
+                int_stop_try2then3args = Try2Then3Args(int_stop)
+
+                def int_step_callback(t, y, H):
+                    pt = qarray(y.reshape(self._d, -1))
+                    return int_stop_try2then3args(t, pt, H)
+
+            # else if there is neither kind of callback but a progbar is
+            #   needed, set up a dummy callback so it gets updated
+            elif self._progbar:
+
+                def int_step_callback(t, y, H):
+                    pass
+
+            # else there are no callbacks and no progbar
+            else:
+                int_step_callback = None
+
+        # else there is compute callback, but may need to add int_stop callback
         else:
-            self._results = []
+            # if both kinds of callback, combine them
+            if int_stop is not None:
+                int_stop_try2then3args = Try2Then3Args(int_stop)
 
-            @functools.wraps(fn)
-            def step_callback(t, pt):
-                self._results.append(fn(t, pt))
+                def int_step_callback(t, y, H):
+                    # For the integration callback, additionally need to
+                    #   convert back to 'quantum' (column vector) form
+                    pt = qarray(y.reshape(self._d, -1))
+                    step_callback(t, pt, H)
+                    return int_stop_try2then3args(t, pt, H)
 
-            @functools.wraps(fn)
-            def int_step_callback(t, y):
-                pt = qarray(y.reshape(self._d, -1))
-                self._results.append(fn(t, pt))
+            # else no int_stop callback, so just set up compute callback
+            else:
+
+                def int_step_callback(t, y, H):
+                    # For the integration callback, additionally need to
+                    #   convert back to 'quantum' (column vector) form
+                    pt = qarray(y.reshape(self._d, -1))
+                    step_callback(t, pt, H)
 
         self._step_callback = step_callback
         self._int_step_callback = int_step_callback
 
-    def _solve_ham(self, ham):
-        """Solve the supplied hamiltonian and find the initial state in the
-        energy eigenbasis for quick evolution later.
+    def _setup_solved_ham(self):
+        """Solve the hamiltonian if needed and find the initial state
+        in the energy eigenbasis for quick evolution later.
         """
         # See if already solved from tuple
         try:
-            self._evals, self._evecs = ham
-            self._method = 'solve'
+            evals, evecs = self._ham
+            self._method = "solve"
         except ValueError:
-            self._evals, self._evecs = eigh(ham.A)
+            evals, evecs = eigh(self._ham.toarray())
+            self._ham = (evals, evecs)
 
         # Find initial state in energy eigenbasis at t0
         if self._isdop:
-            self.pe0 = dot(dag(self._evecs), dot(self._p0, self._evecs))
+            self.pe0 = dot(dag(evecs), dot(self._p0, evecs))
             self._update_method = self._update_to_solved_dop
         else:
-            self.pe0 = dot(dag(self._evecs), self._p0)
+            self.pe0 = dot(dag(evecs), self._p0)
             self._update_method = self._update_to_solved_ket
 
         # Current state (start with same as initial)
         self._pt = self._p0
-        self._solved = True
 
     def _start_integrator(self, ham, small_step):
-        """Initialize a stepping integrator.
-        """
-        self._sparse_ham = issparse(ham)
+        """Initialize a stepping integrator."""
+        if self._timedep:
+            H0 = ham(0.0)
+        else:
+            H0 = ham
 
         # set complex ode with governing equation
-        evo_eq = _calc_evo_eq(self._isdop, self._sparse_ham)
+        evo_eq = _calc_evo_eq(self._isdop, issparse(H0), False, self._timedep)
+
         self._stepper = complex_ode(evo_eq(ham))
 
         # 5th order stpper or 8th order stepper
-        int_mthd, step_fct = ('dopri5', 150) if small_step else ('dop853', 50)
-        first_step = norm(ham, 'f') / step_fct
+        int_mthd, step_fct = ("dopri5", 150) if small_step else ("dop853", 50)
+        if isinstance(H0, LinearOperator):
+            # approx norm doesn't need to be very accurate
+            nrm0 = norm_fro_approx(H0, tol=0.1)
+        else:
+            nrm0 = norm(H0, "f")
+        first_step = nrm0 / step_fct
 
         self._stepper.set_integrator(int_mthd, nsteps=0, first_step=first_step)
 
         # Set step_callback to be evaluated with args (t, y) at each step
         if self._int_step_callback is not None:
-            self._stepper.set_solout(self._int_step_callback)
 
-        self._stepper.set_initial_value(self._p0.A.reshape(-1), self.t0)
+            def solout(t, y):
+                res = self._int_step_callback(t, y, self._ham)
+                return res
+
+            self._stepper.set_solout(solout)
+
+        self._stepper.set_initial_value(
+            self._p0.toarray().reshape(-1), self.t0
+        )
 
         # assign the correct update_to method
         self._update_method = self._update_to_integrate
-        self._solved = False
 
     # Methods for updating the simulation ----------------------------------- #
 
@@ -375,42 +592,47 @@ class Evolution(object):
         the operator exponential itself.
         """
         factor = -1j * (t - self.t)
-        self._pt = expm_multiply(factor * self.ham, self._pt,
-                                 backend=self.expm_backend, **self.expm_opts)
+        self._pt = expm_multiply(
+            factor * self._ham,
+            self._pt,
+            backend=self.expm_backend,
+            **self.expm_opts,
+        )
         self._t = t
 
         # compute any callbacks into -> self._results
         if self._step_callback is not None:
-            self._step_callback(t, self._pt)
+            self._step_callback(t, self._pt, self._ham)
 
     def _update_to_solved_ket(self, t):
         """Update simulation consisting of a solved hamiltonian and a
         wavefunction to time `t`.
         """
         self._t = t
-        lt = explt(self._evals, t - self.t0)
-        self._pt = self._evecs @ ldmul(lt, self.pe0)
+        evals, evecs = self._ham
+        lt = explt(evals, t - self.t0)
+        self._pt = evecs @ ldmul(lt, self.pe0)
 
         # compute any callbacks into -> self._results
         if self._step_callback is not None:
-            self._step_callback(t, self._pt)
+            self._step_callback(t, self._pt, self._ham)
 
     def _update_to_solved_dop(self, t):
         """Update simulation consisting of a solved hamiltonian and a
         density operator to time `t`.
         """
         self._t = t
-        lt = explt(self._evals, t - self.t0)
+        evals, evecs = self._ham
+        lt = explt(evals, t - self.t0)
         lvpvl = rdmul(ldmul(lt, self.pe0), lt.conj())
-        self._pt = self._evecs @ (lvpvl @ dag(self._evecs))
+        self._pt = evecs @ (lvpvl @ dag(evecs))
 
         # compute any callbacks into -> self._results
         if self._step_callback is not None:
-            self._step_callback(t, self._pt)
+            self._step_callback(t, self._pt, self._ham)
 
     def _update_to_integrate(self, t):
-        """Update simulation consisting of unsolved hamiltonian.
-        """
+        """Update simulation consisting of unsolved hamiltonian."""
         self._stepper.integrate(t)
 
     def update_to(self, t):
@@ -421,14 +643,16 @@ class Evolution(object):
         t : float
             Time to update the evolution to.
         """
-        if self._progbar and hasattr(self, '_stepper'):
-
+        if self._progbar and hasattr(self, "_stepper"):
             with continuous_progbar(self.t, t) as pbar:
                 if self._int_step_callback is not None:
+
                     def solout(t, y):
-                        self._int_step_callback(t, y)
+                        int_stop_res = self._int_step_callback(t, y, self._ham)
                         pbar.cupdate(t)
+                        return int_stop_res
                 else:
+
                     def solout(t, _):
                         pbar.cupdate(t)
 
@@ -469,15 +693,13 @@ class Evolution(object):
 
     @property
     def t(self):
-        """float : Current time of simulation.
-        """
-        return self._stepper.t if self._method == 'integrate' else self._t
+        """float : Current time of simulation."""
+        return self._stepper.t if self._method == "integrate" else self._t
 
     @property
     def pt(self):
-        """quantum state : State of the system at the current time (t).
-        """
-        if self._method == 'integrate':
+        """quantum state : State of the system at the current time (t)."""
+        if self._method == "integrate":
             return qarray(self._stepper.y.reshape(self._d, -1))
         else:
             return self._pt
