@@ -1,16 +1,15 @@
 """Functions for generating quantum operators.
 """
-from operator import add
 import math
 import functools
 import itertools
 import operator
 
-from cytoolz import isiterable, concat, unique
 import numpy as np
 import scipy.sparse as sp
 from scipy.special import comb
 
+from ..utils import isiterable, concat, unique
 from ..core import (qarray, make_immutable, get_thread_pool,
                     par_reduce, isreal, qu, eye, kron, ikron)
 
@@ -27,12 +26,14 @@ def spin_operator(label, S=1 / 2, **kwargs):
     ----------
     label : str
         The type of operator, can be one of six options:
+
             - ``{'x', 'X'}``, x-spin operator.
             - ``{'y', 'Y'}``, y-spin operator.
             - ``{'z', 'Z'}``, z-spin operator.
             - ``{'+', 'p'}``, Raising operator.
             - ``{'-', 'm'}``, Lowering operator.
             - ``{'i', 'I'}``, identity operator.
+
     S : float, optional
         The spin of particle to act on, default to spin-1/2.
     kwargs
@@ -93,8 +94,8 @@ def spin_operator(label, S=1 / 2, **kwargs):
         np.fill_diagonal(op, 1.0)
 
     else:
-        raise ValueError("Label '{}'' not understood, should be one of ``['X',"
-                         " 'Y', 'Z', '+', '-', 'I']``.".format(label))
+        raise ValueError(f"Label '{label}'' not understood, should be one of "
+                         "``['X', 'Y', 'Z', '+', '-', 'I']``.")
 
     op = qu(np.real_if_close(op), **kwargs)
 
@@ -205,6 +206,108 @@ Ry = functools.partial(rotation, xyz='y')
 Rz = functools.partial(rotation, xyz='z')
 
 
+@functools.lru_cache(128)
+def U_gate(theta, phi, lamda, dtype=complex, sparse=False):
+    r"""Arbitrary unitary single qubit gate.
+
+    .. math::
+
+        U_3(\theta, \phi, \lambda) =
+        \begin{bmatrix}
+        \cos(\theta / 2) & - e^{i \lambda} \sin(\theta / 2) \\
+        e^{i \phi} \sin(\theta / 2) & e^{i(\lambda + \phi)}\cos(\theta / 2)
+        \end{bmatrix}
+
+    Parameters
+    ----------
+    theta : float
+        Angle between 0 and pi.
+    phi : float
+        Angle between 0 and 2 pi.
+    lamba : float
+        Angle between 0 and 2 pi.
+
+    Returns
+    -------
+    U : (2, 2) array
+        The unitary matrix, cached.
+    """
+    from cmath import cos, sin, exp
+
+    c2, s2 = cos(theta / 2), sin(theta / 2)
+    U = qu(
+        [[c2, -exp(1j * lamda) * s2],
+         [exp(1j * phi) * s2, exp(1j * (lamda + phi)) * c2]],
+        dtype=dtype, sparse=sparse
+    )
+    make_immutable(U)
+    return U
+
+
+@functools.lru_cache(4)
+def Xsqrt(**qu_opts):
+    r"""Rx(pi / 2).
+
+    .. math::
+
+        X^{\frac{1}{2}} =
+        \frac{1}{\sqrt{2}}
+        \begin{bmatrix}
+        1  & - i \\
+        - i & 1
+        \end{bmatrix}
+    """
+    return Rx(math.pi / 2, **qu_opts)
+
+
+@functools.lru_cache(4)
+def Ysqrt(**qu_opts):
+    r"""Ry(pi / 2).
+
+    .. math::
+
+        Y^{\frac{1}{2}} =
+        \frac{1}{\sqrt{2}}
+        \begin{bmatrix}
+        1 & - 1 \\
+        1 & 1
+        \end{bmatrix}
+    """
+    return Ry(math.pi / 2, **qu_opts)
+
+
+@functools.lru_cache(4)
+def Zsqrt(**qu_opts):
+    r"""Rz(pi / 2).
+
+    .. math::
+
+        Z^{\frac{1}{2}} =
+        \frac{1}{\sqrt{2}}
+        \begin{bmatrix}
+        1 - i & 0 \\
+        0 & 1 + i
+        \end{bmatrix}
+    """
+    return Rz(math.pi / 2, **qu_opts)
+
+
+@functools.lru_cache(4)
+def Wsqrt(**qu_opts):
+    r"""R[X + Y](pi / 2).
+
+    .. math::
+
+        W^{\frac{1}{2}} =
+        \frac{1}{\sqrt{2}}
+        \begin{bmatrix}
+        1 & -\sqrt{i} \\
+        \sqrt{-i} & 1
+        \end{bmatrix}
+    """
+    return U_gate(math.pi / 2, -math.pi / 4, math.pi / 4, **qu_opts)
+
+
 @functools.lru_cache(maxsize=8)
 def swap(dim=2, dtype=complex, **kwargs):
     """The SWAP operator acting on subsystems of dimension `dim`.
@@ -218,6 +321,122 @@ def swap(dim=2, dtype=complex, **kwargs):
     return S
 
 
+@functools.lru_cache(maxsize=128)
+def fsim(theta, phi, dtype=complex, **kwargs):
+    r"""The 'fermionic simulation' gate:
+
+    .. math::
+
+        \mathrm{fsim}(\theta, \phi) =
+        \begin{bmatrix}
+        1 & 0 & 0 & 0\\
+        0 & \cos(\theta) & -i sin(\theta) & 0\\
+        0 & -i sin(\theta) & \cos(\theta) & 0\\
+        0 & 0 & 0 & \exp(-i \phi)
+        \end{bmatrix}
+
+    Note that ``theta`` and ``phi`` should be specified in radians and the sign
+    convention with this gate varies. Here for example,
+    ``fsim(- pi / 2, 0) == iswap()``.
+    """
+    from cmath import cos, sin, exp
+
+    a = cos(theta)
+    b = -1j * sin(theta)
+    c = exp(-1j * phi)
+    gate = [[1, 0, 0, 0],
+            [0, a, b, 0],
+            [0, b, a, 0],
+            [0, 0, 0, c]]
+
+    gate = qu(gate, dtype=dtype, **kwargs)
+    make_immutable(gate)
+    return gate
+
+
+@functools.lru_cache(maxsize=256)
+def fsimg(theta, zeta, chi, gamma, phi, dtype=complex, **kwargs):
+    r"""The 'fermionic simulation' gate, with:
+
+        * :math:`\theta` is the iSWAP angle
+        * :math:`\phi` is the controlled-phase angle
+        * :math:`\zeta, \chi, \gamma`  are single-qubit phase angles.
+
+    .. math::
+        \mathrm{fsimg}(\theta, \zeta, \chi, \gamma, \phi) =
+        \begin{bmatrix}
+        1 & 0 & 0 & 0\\
+        0 & \exp(-i(\gamma +\zeta )) \cos(\theta) &
+        -i \exp(-i(\gamma - \chi )) sin(\theta) & 0\\
+        0 & -i \exp(-i(\gamma + \chi )) sin(\theta) &
+        \exp(-i(\gamma - \zeta )) \cos(\theta) & 0\\
+        0 & 0 & 0 & \exp(-i (\phi +2 \gamma))
+        \end{bmatrix}
+
+    See Equation 18 of https://arxiv.org/abs/2010.07965. Note that ``theta``,
+    ``phi``, ``zeta``, ``chi``, ``gamma`` should be specified in radians and
+    the sign convention with this gate varies. Here for example,
+    ``fsimg(- pi / 2, 0, 0, 0,0) == iswap()``.
+    """
+    from cmath import cos, sin, exp
+
+    a1 = exp(-1j * (gamma + zeta)) * cos(theta)
+    a2 = exp(-1j * (gamma - zeta)) * cos(theta)
+
+    b1 = -1j * exp(-1j * (gamma - chi)) * sin(theta)
+    b2 = -1j * exp(-1j * (gamma + chi)) * sin(theta)
+
+    c = exp(-1j * (phi + 2 * gamma))
+
+    gate = [[1, 0, 0, 0],
+            [0, a1, b1, 0],
+            [0, b2, a2, 0],
+            [0, 0, 0, c]]
+    gate = qu(gate, dtype=dtype, **kwargs)
+    make_immutable(gate)
+    return gate
+
+
+@functools.lru_cache(maxsize=4)
+def iswap(dtype=complex, **kwargs):
+    iswap = qu([[1., 0., 0., 0.],
+                [0., 0., 1j, 0.],
+                [0., 1j, 0., 0.],
+                [0., 0., 0., 1.]], dtype=dtype, **kwargs)
+    make_immutable(iswap)
+    return iswap
+
+
+def ncontrolled_gate(ncontrol, gate, dtype=complex, sparse=False):
+    """Build an n-qubit controlled gate. The control qubits are the
+    first ``ncontrol`` qubits.
+
+    Parameters
+    ----------
+    ncontrol : int
+        The number of control qubits.
+    gate : array_like
+        The gate to apply to the controlled qubit(s).
+    dtype : str
+        The data type of the returned matrix.
+    sparse : bool
+        Whether to return a sparse matrix.
+
+    Returns
+    -------
+    C : qarray
+        The n-qubit controlled gate.
+    """
+    dG = gate.shape[0]
+    d = 2**ncontrol * dG
+    # build gate dense and dtype='complex128'
+    op = np.identity(d, dtype='complex128')
+    op[-dG:, -dG:] = gate
+    # then convert to desired dtype and format
+    op = qu(op, dtype=dtype, sparse=sparse)
+    return op
+
+
 @functools.lru_cache(maxsize=16)
 def controlled(s, dtype=complex, sparse=False):
     """Construct a controlled pauli gate for two qubits.
@@ -231,14 +450,13 @@ def controlled(s, dtype=complex, sparse=False):
 
     Returns
     -------
-    C : immutable operator
+    C : qarray
         The controlled two-qubit gate operator.
     """
     # alias not and NOT to x
     s = {'NOT': 'x', 'not': 'x'}.get(s, s)
-    kws = {'dtype': dtype, 'sparse': sparse}
-    op = ((qu([1, 0], qtype='dop', **kws) & eye(2, **kws)) +
-          (qu([0, 1], qtype='dop', **kws) & pauli(s, **kws)))
+    gate = pauli(s)
+    op = ncontrolled_gate(1, gate, dtype=dtype, sparse=sparse)
     make_immutable(op)
     return op
 
@@ -271,6 +489,48 @@ def cZ(dtype=complex, sparse=False):
     return controlled('Z', dtype=dtype, sparse=sparse)
 
 
+@functools.lru_cache(8)
+def ccX(dtype=complex, sparse=False):
+    """The double controlled X gate, or Toffoli gate.
+    """
+    op = ncontrolled_gate(2, pauli('X'), dtype=dtype, sparse=sparse)
+    make_immutable(op)
+    return op
+
+
+@functools.lru_cache(8)
+def ccY(dtype=complex, sparse=False):
+    """The double controlled Y gate.
+    """
+    op = ncontrolled_gate(2, pauli('Y'), dtype=dtype, sparse=sparse)
+    make_immutable(op)
+    return op
+
+
+@functools.lru_cache(8)
+def ccZ(dtype=complex, sparse=False):
+    """The double controlled Z gate.
+    """
+    op = ncontrolled_gate(2, pauli('Z'), dtype=dtype, sparse=sparse)
+    make_immutable(op)
+    return op
+
+
+@functools.lru_cache(8)
+def controlled_swap(dtype=complex, sparse=False):
+    """The controlled swap or Fredkin gate. The control qubit is the first
+    qubit, if in state |1> a swap is performed on the last two qubits.
+    """
+    op = ncontrolled_gate(1, swap(), dtype=dtype, sparse=sparse)
+    make_immutable(op)
+    return op
+
+
+cswap = controlled_swap
+fredkin = controlled_swap
+toffoli = ccX
+
+
 # --------------------------------------------------------------------------- #
 #                                Hamiltonians                                 #
 # --------------------------------------------------------------------------- #
@@ -281,7 +541,8 @@ def hamiltonian_builder(fn):
     kwargs ``stype`` and ``sparse``. This assumes the core function always
     builds the hamiltonian in sparse form. The wrapper then:
 
-    1. Checks if the operator is real
+    1. Checks if the operator is real and, if so, discards imaginary part if no
+       explicity `dtype` was given
     2. Converts the operator to dense or the correct sparse form
     3. Makes the operator immutable so it can be safely cached
     """
@@ -290,11 +551,11 @@ def hamiltonian_builder(fn):
     def ham_fn(*args, stype='csr', sparse=False, **kwargs):
         H = fn(*args, **kwargs)
 
-        if isreal(H):
+        if kwargs.get('dtype', None) is None and isreal(H):
             H = H.real
 
         if not sparse:
-            H = qarray(H.A)
+            H = qarray(H.toarray())
         elif H.format != stype:
             H = H.asformat(stype)
 
@@ -307,8 +568,8 @@ def hamiltonian_builder(fn):
 
 @functools.lru_cache(maxsize=8)
 @hamiltonian_builder
-def ham_heis(n, j=1.0, b=0.0, cyclic=True,
-             parallel=None, nthreads=None, ownership=None):
+def ham_heis(n, j=1.0, b=0.0, cyclic=False,
+             parallel=False, nthreads=None, ownership=None):
     """Constructs the nearest neighbour 1d heisenberg spin-1/2 hamiltonian.
 
     Parameters
@@ -387,12 +648,12 @@ def ham_heis(n, j=1.0, b=0.0, cyclic=True,
         # General term, on-site b-field plus interaction with next site
         return ikron(two_site_term, dims, [i, i + 1], **ikron_kws)
 
-    terms_needed = range(0 if single_site_b is 0 else -1,
+    terms_needed = range(0 if not any((bx, by, bz)) else -1,
                          n if cyclic else n - 1)
 
     if parallel:
         pool = get_thread_pool(nthreads)
-        ham = par_reduce(add, pool.map(gen_term, terms_needed))
+        ham = par_reduce(operator.add, pool.map(gen_term, terms_needed))
     else:
         ham = sum(map(gen_term, terms_needed))
 
@@ -425,7 +686,7 @@ def ham_XXZ(n, delta, jxy=1.0, **ham_opts):
 
 @functools.lru_cache(maxsize=8)
 @hamiltonian_builder
-def ham_j1j2(n, j1=1.0, j2=0.5, bz=0.0, cyclic=True, ownership=None):
+def ham_j1j2(n, j1=1.0, j2=0.5, bz=0.0, cyclic=False, ownership=None):
     """Generate the j1-j2 hamiltonian, i.e. next nearest neighbour
     interactions.
 
@@ -513,7 +774,7 @@ def _gen_mbl_random_factors(n, dh, dh_dim, dh_dist, seed=None, beta=None):
         rs = 2.0 * np.random.rand(3, n) - 1.0
 
     elif dh_dist in {'qp', 'quasiperiodic', 'qr', 'quasirandom'}:
-        if dh_dim is not 'z':
+        if dh_dim != 'z':
             raise ValueError("dh_dim should be 1 or 'z' for dh_dist='qp'.")
 
         if beta is None:
@@ -531,7 +792,7 @@ def _gen_mbl_random_factors(n, dh, dh_dim, dh_dist, seed=None, beta=None):
 
 
 @hamiltonian_builder
-def ham_mbl(n, dh, j=1.0, bz=0.0, cyclic=True,
+def ham_mbl(n, dh, j=1.0, bz=0.0, cyclic=False,
             seed=None, dh_dist="s", dh_dim=1, beta=None, ownership=None):
     """ Constructs a heisenberg hamiltonian with isotropic coupling and
     random fields acting on each spin - the many-body localized (MBL)
@@ -612,7 +873,16 @@ def ham_mbl(n, dh, j=1.0, bz=0.0, cyclic=True,
 @hamiltonian_builder
 def ham_heis_2D(n, m, j=1.0, bz=0.0, cyclic=False,
                 parallel=False, ownership=None):
-    """Construct the 2D spin-1/2 heisenberg model hamiltonian.
+    r"""Construct the 2D spin-1/2 heisenberg model hamiltonian:
+
+    .. math::
+
+        \hat{H} = \sum_{<i, j>}
+        J_X S^X_i S^X_j +
+        J_Y S^Y_i S^Y_j +
+        J_Z S^Z_i S^Z_j
+
+    where the sum runs over pairs :math:`<i,j>` on a 2D square lattice.
 
     Parameters
     ----------
@@ -651,6 +921,8 @@ def ham_heis_2D(n, m, j=1.0, bz=0.0, cyclic=False,
     except (TypeError, ValueError):
         jx = jy = jz = j
 
+    js = {s: js for s, js in zip("xyz", [jx, jy, jz]) if js != 0.0}
+
     dims = [[2] * m] * n  # shape (n, m)
 
     sites = tuple(itertools.product(range(n), range(m)))
@@ -665,21 +937,20 @@ def ham_heis_2D(n, m, j=1.0, bz=0.0, cyclic=False,
             if cyclic or right != 0:
                 yield ((i, j), (i, right))
 
+    # generate all pairs of coordinates and directions
+    pairs_ss = tuple(itertools.product(gen_pairs(), js))
+
     # build the hamiltonian in sparse 'coo' format always for efficiency
     op_kws = {'sparse': True, 'stype': 'coo'}
     ikron_kws = {'sparse': True, 'stype': 'coo',
                  'coo_build': True, 'ownership': ownership}
 
-    # generate all pairs of coordinates and directions
-    pairs_ss = tuple(itertools.product(gen_pairs(), 'xyz'))
-
     # generate XX, YY and ZZ interaction from
     #     e.g. arg ([(3, 4), (3, 5)], 'z')
     def interactions(pair_s):
         pair, s = pair_s
         Sxyz = spin_operator(s, **op_kws)
-        J = {'x': jx, 'y': jy, 'z': jz}[s]
-        return ikron(J * Sxyz, dims, inds=pair, **ikron_kws)
+        return ikron([js[s] * Sxyz, Sxyz], dims, inds=pair, **ikron_kws)
 
     # generate Z field
     def fields(site):
@@ -691,13 +962,13 @@ def ham_heis_2D(n, m, j=1.0, bz=0.0, cyclic=False,
         all_terms = itertools.chain(
             map(interactions, pairs_ss),
             map(fields, sites) if bz != 0.0 else ())
-        H = sum(all_terms)
+        H = functools.reduce(operator.add, all_terms)
     else:
         pool = get_thread_pool()
         all_terms = itertools.chain(
             pool.map(interactions, pairs_ss),
             pool.map(fields, sites) if bz != 0.0 else ())
-        H = par_reduce(add, all_terms)
+        H = par_reduce(operator.add, all_terms)
 
     return H
 
@@ -750,7 +1021,7 @@ def zspin_projector(n, sz=0, stype="csr", dtype=float):
 
     Examples
     --------
-    >>> zspin_projector(n=2, sz=0).A
+    >>> zspin_projector(n=2, sz=0).toarray()
     array([[0., 0.],
            [1., 0.],
            [0., 1.],
@@ -777,8 +1048,8 @@ def zspin_projector(n, sz=0, stype="csr", dtype=float):
         # Number of 'up' spins
         k = n / 2 + s
         if not k.is_integer():
-            raise ValueError("{} is not a valid spin half subspace for "
-                             "{} spins.".format(s, n))
+            raise ValueError(f"{s} is not a valid spin half subspace for {n} "
+                             "spins.")
         k = int(round(k))
         # Size of subspace
         p += comb(n, k, exact=True)
@@ -833,8 +1104,8 @@ def num(n, **qu_opts):
 
 @functools.lru_cache(maxsize=8)
 @hamiltonian_builder
-def ham_hubbard_hardcore(n, t=0.5, V=1., mu=1., cyclic=True,
-                         parallel=None, ownership=None):
+def ham_hubbard_hardcore(n, t=0.5, V=1., mu=1., cyclic=False,
+                         parallel=False, ownership=None):
     """Generate the spinless fermion hopping hamiltonian.
 
     Parameters
